import httpx
from typing import Any, List

from textual.app import ComposeResult
from textual.containers import VerticalScroll
from textual.reactive import reactive
from textual.widgets import Button, Static

from llama_agents.app.components.types import ButtonType


class TaskButton(Button):
    type: ButtonType = ButtonType.TASK


class TasksList(Static):
    tasks: List[str] = reactive([])

    def __init__(self, control_plane_url: str, **kwargs: Any):
        self.control_plane_url = control_plane_url
        super().__init__(**kwargs)

    def compose(self) -> ComposeResult:
        with VerticalScroll(id="tasks-scroll"):
            for task in self.tasks:
                yield TaskButton(task)

    async def on_mount(self) -> None:
        self.set_interval(5, self.refresh_tasks)

    async def refresh_tasks(self) -> None:
        async with httpx.AsyncClient(timeout=120.0) as client:
            response = await client.get(f"{self.control_plane_url}/tasks")
            tasks_dict = response.json()

        new_tasks = []
        for task_id in tasks_dict:
            new_tasks.append(task_id)

        self.tasks = [*new_tasks]

    async def watch_tasks(self, new_tasks: List[str]) -> None:
        try:
            tasks_scroll = self.query_one("#tasks-scroll")
            await tasks_scroll.remove_children()
            for task in new_tasks:
                await tasks_scroll.mount(TaskButton(task))
        except Exception:
            pass



import argparse

from llama_agents.app.app import run as launch_monitor


def main() -> None:
    parser = argparse.ArgumentParser(description="llama-agents CLI interface.")

    # Subparsers for the main commands
    subparsers = parser.add_subparsers(title="commands", dest="command", required=True)

    # Subparser for the monitor command
    monitor_parser = subparsers.add_parser("monitor", help="Monitor the agents.")
    monitor_parser.add_argument(
        "--control-plane-url",
        default="http://127.0.0.1:8000",
        help="The URL of the control plane. Defaults to http://127.0.0.1:8000",
    )
    monitor_parser.set_defaults(
        func=lambda args: launch_monitor(args.control_plane_url)
    )

    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    main()



import asyncio
import signal
import sys
import uuid
from typing import Any, Callable, List, Optional

from llama_agents.services.base import BaseService
from llama_agents.control_plane.base import BaseControlPlane
from llama_agents.message_consumers.base import (
    BaseMessageQueueConsumer,
    StartConsumingCallable,
)
from llama_agents.message_queues.simple import SimpleMessageQueue
from llama_agents.message_queues.base import PublishCallback, BaseMessageQueue
from llama_agents.message_publishers.publisher import MessageQueuePublisherMixin


class ServerLauncher(MessageQueuePublisherMixin):
    """Launches a llama-agents system as a server.

    The ServerLauncher is a convenience class for launching a system of services as a server.

    When launching, the launcher will:
    - Register each service to the control plane.
    - Start each service.
    - Register any additional consumers (like handling the end of a task with a `human` consumer).
    - Run until the system is shut down.

    Args:
        services (List[BaseService]): List of services to launch.
        control_plane (BaseControlPlane): Control plane for the system.
        message_queue (SimpleMessageQueue): Message queue for the system.
        publish_callback (Optional[PublishCallback], optional): Callback for publishing messages. Defaults to None.
        additional_consumers (Optional[List[BaseMessageQueueConsumer]], optional): Additional consumers to register. Defaults to None.

    Examples:
        One script might launch the servers:
        ```python
        from llama_agents import ServerLauncher

        launcher = ServerLauncher(
            services=[service1, service2],
            control_plane=control_plane,
            message_queue=message_queue,
        )
        launcher.launch_servers()
        ```
        Another script would interact with the system using the client:
        ```python
        import time
        from llama_agents import LlamaAgentsClient

        client = LlamaAgentsClient("<control_plane_url>")

        # interact with the system
        task_id = client.create_task("Do the thing.")
        time.sleep(15) # wait for the task to complete!
        result = client.get_task_result(task_id)
        ```
    """

    def __init__(
        self,
        services: List[BaseService],
        control_plane: BaseControlPlane,
        message_queue: BaseMessageQueue,
        publish_callback: Optional[PublishCallback] = None,
        additional_consumers: Optional[List[BaseMessageQueueConsumer]] = None,
    ) -> None:
        self.services = services
        self.control_plane = control_plane
        self._message_queue = message_queue or SimpleMessageQueue()
        self._publisher_id = f"{self.__class__.__qualname__}-{uuid.uuid4()}"
        self._publish_callback = publish_callback
        self._additional_consumers = additional_consumers or []

    @property
    def message_queue(self) -> SimpleMessageQueue:
        return self._message_queue

    @property
    def publisher_id(self) -> str:
        return self._publisher_id

    @property
    def publish_callback(self) -> Optional[PublishCallback]:
        return self._publish_callback

    def get_shutdown_handler(self, tasks: List[asyncio.Task]) -> Callable:
        def signal_handler(sig: Any, frame: Any) -> None:
            print("\nShutting down.")
            for task in tasks:
                task.cancel()
            sys.exit(0)

        return signal_handler

    def launch_servers(self) -> None:
        """Launch the system in multiple FastAPI servers."""
        return asyncio.run(self.alaunch_servers())

    async def alaunch_servers(self) -> None:
        """Launch the system in multiple FastAPI servers."""
        # launch the message queue
        queue_task = asyncio.create_task(self.message_queue.launch_server())

        # wait for the message queue to be ready
        await asyncio.sleep(1)

        # launch the control plane
        control_plane_task = asyncio.create_task(self.control_plane.launch_server())

        # wait for the control plane to be ready
        await asyncio.sleep(1)

        # register the control plane as a consumer
        start_consuming_callables: List[StartConsumingCallable] = []
        start_consuming_callables.append(
            await self.control_plane.register_to_message_queue()
        )

        # register the services
        control_plane_url = f"http://{self.control_plane.host}:{self.control_plane.port}"  # type: ignore
        service_tasks: List[asyncio.Task] = []
        for service in self.services:
            service_tasks.append(asyncio.create_task(service.launch_server()))
            start_consuming_callables.append(await service.register_to_message_queue())
            await service.register_to_control_plane(control_plane_url)

        # register additional consumers
        for consumer in self._additional_consumers:
            start_consuming_callables.append(
                await self.message_queue.register_consumer(consumer)
            )

        # consumers start consuming in their own threads
        start_consuming_tasks = []
        for start_consuming_callable in start_consuming_callables:
            task: asyncio.Task = asyncio.create_task(start_consuming_callable())
            start_consuming_tasks.append(task)

        shutdown_handler = self.get_shutdown_handler(
            [*service_tasks, *start_consuming_tasks, queue_task, control_plane_task]
        )
        loop = asyncio.get_event_loop()
        while loop.is_running():
            await asyncio.sleep(0.1)
            signal.signal(signal.SIGINT, shutdown_handler)

            for task in [
                *service_tasks,
                *start_consuming_tasks,
                queue_task,
                control_plane_task,
            ]:
                if task.done() and task.exception():  # type: ignore
                    raise task.exception()  # type: ignore



from abc import ABC, abstractmethod
from typing import Any, Dict, List, Tuple

from llama_index.core.tools import BaseTool

from llama_agents.messages.base import QueueMessage
from llama_agents.types import TaskDefinition, TaskResult


class BaseOrchestrator(ABC):
    """Base class for an orchestrator.

    The general idea for an orchestrator is to manage the flow of messages between services.

    Given some state, task, and list of tools, figure out the next messages to publish. Then, once
    the messages are processed, update the state with the results.

    Currently, the final message is published to the `human` message queue for final processing.
    """

    @abstractmethod
    async def get_next_messages(
        self, task_def: TaskDefinition, tools: List[BaseTool], state: Dict[str, Any]
    ) -> Tuple[List[QueueMessage], Dict[str, Any]]:
        """Get the next message to process. Returns the message and the new state."""
        ...

    @abstractmethod
    async def add_result_to_state(
        self, result: TaskResult, state: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Add the result of processing a message to the state. Returns the new state."""
        ...



import json
import pickle
from typing import Any, Dict, List, Tuple

from llama_index.core.query_pipeline import QueryPipeline
from llama_index.core.query_pipeline.query import RunState
from llama_index.core.tools import BaseTool

from llama_agents.messages.base import QueueMessage
from llama_agents.orchestrators.base import BaseOrchestrator
from llama_agents.tools.service_component import ServiceComponent, ModuleType
from llama_agents.types import ActionTypes, TaskDefinition, TaskResult

RUN_STATE_KEY = "run_state"
NEXT_SERVICE_KEYS = "next_service_keys"
LAST_MODULES_RUN = "last_modules_run"
RESULT_KEY = "result"


INPUT_DICT_KEY = "__input_dict__"


def get_service_component_message(
    module: ServiceComponent,
    task_id: str,
    input_dict: Dict[str, Any],
) -> QueueMessage:
    """Enqueue a service component message.

    This depends directly on whether the service component
    wraps an agent service or a component

    """
    if module.module_type == ModuleType.AGENT:
        # in an agent, assume input_dict is a single input
        input = next(iter(input_dict.values()))
        queue_message = QueueMessage(
            type=module.name,
            action=ActionTypes.NEW_TASK,
            data=TaskDefinition(
                input=input,
                task_id=task_id,
            ).model_dump(),
        )
    elif module.module_type == ModuleType.COMPONENT:
        # in a component, input_dict is a dict
        queue_message = QueueMessage(
            type=module.name,
            action=ActionTypes.NEW_TASK,
            data=TaskDefinition(
                input="",
                task_id=task_id,
                state={"__input_dict__": input_dict} if input_dict else {},
            ).model_dump(),
        )
    else:
        raise ValueError("Invalid module type")

    return queue_message


def process_component_output(
    pipeline: QueryPipeline,
    run_state: RunState,
    module_key: str,
    task_result: TaskResult,
) -> None:
    """Process component outputs.

    Take the task result and update the next modules in the pipeline.

    Propagate different data depending on the module type.

    """
    module = run_state.module_dict[module_key]

    # assume an agent component if the module type is not set
    if not hasattr(module, "module_type") or module.module_type == ModuleType.AGENT:
        # in an agent, the output is a single value
        pipeline.process_component_output(
            {"output": task_result.result},
            module_key,
            run_state,
        )
        return

    elif module.module_type == ModuleType.COMPONENT:
        # in a component, the output is a dict
        pipeline.process_component_output(task_result.data, module_key, run_state)

    else:
        raise ValueError("Invalid module type")


class PipelineOrchestrator(BaseOrchestrator):
    """Orchestrator for a query pipeline.

    Given an incoming task, process it through a query pipeline that may contain
    calls to external `llama-agents` services.

    Attributes:
        pipeline (QueryPipeline): The query pipeline to run.

    Examples:
        ```python
        from llama_index.core.query_pipeline import QueryPipeline
        from llama_agents import PipelineOrchestrator, AgentService, ServiceComponent

        query_rewrite_server = AgentService(
            agent=hyde_agent,
            message_queue=message_queue,
            description="Used to rewrite queries",
            service_name="query_rewrite_agent",
            host="127.0.0.1",
            port=8011,
        )
        query_rewrite_server_c = ServiceComponent.from_service_definition(query_rewrite_server)

        rag_agent_server = AgentService(
            agent=rag_agent,
            message_queue=message_queue,
            description="rag_agent",
            host="127.0.0.1",
            port=8012,
        )
        rag_agent_server_c = ServiceComponent.from_service_definition(rag_agent_server)

        # create our multi-agent framework components
        pipeline = QueryPipeline(chain=[query_rewrite_server_c, rag_agent_server_c])
        orchestrator = PipelineOrchestrator(pipeline=pipeline)
        ```
    """

    def __init__(
        self,
        pipeline: QueryPipeline,
    ):
        self.pipeline = pipeline

    async def get_next_messages(
        self, task_def: TaskDefinition, tools: List[BaseTool], state: Dict[str, Any]
    ) -> Tuple[List[QueueMessage], Dict[str, Any]]:
        # check if we need to init the state
        if RUN_STATE_KEY not in state:
            run_state = self.pipeline.get_run_state(input=task_def.input)
        else:
            run_state = pickle.loads(state[RUN_STATE_KEY])

        # run the next step in the pipeline, until we hit a service component
        next_module_keys = self.pipeline.get_next_module_keys(run_state)

        next_messages = []
        next_service_keys = []
        found_service_component = False
        module_key = None

        while len(next_module_keys) > 0:
            for module_key in next_module_keys:
                module = run_state.module_dict[module_key]
                module_input = run_state.all_module_inputs[module_key]

                if isinstance(module, ServiceComponent):
                    queue_message = get_service_component_message(
                        module,
                        task_def.task_id,
                        input_dict=module_input,
                    )

                    found_service_component = True
                    next_service_keys.append(module_key)
                    next_messages.append(queue_message)
                    continue

                # run the module if it is not a service component
                output_dict = await module.arun_component(**module_input)

                # check if the output is a service component
                if "service_output" in output_dict:
                    service_dict = json.loads(output_dict["service_output"])
                    found_service_component = True
                    next_service_keys.append(module_key)

                    # if the component isn't a service component, wrap it
                    # this can happen if ServiceComponents are nested in routers
                    new_module = module
                    if not isinstance(module, ServiceComponent):
                        new_module = ServiceComponent(
                            name=service_dict["name"],
                            description=service_dict["description"],
                            module_type=ModuleType.AGENT,
                        )

                    queue_message = get_service_component_message(
                        new_module,
                        task_def.task_id,
                        input_dict=service_dict["input"],
                    )
                    next_messages.append(queue_message)
                    continue

                # process the output if it is not a service component
                self.pipeline.process_component_output(
                    output_dict,
                    module_key,
                    run_state,
                )

            if found_service_component:
                break

            # get the next module keys
            next_module_keys = self.pipeline.get_next_module_keys(
                run_state,
            )

            # if no more modules to run, break
            if len(next_module_keys) == 0:
                run_state.result_outputs[module_key] = output_dict
                break

        # did we find a service component?
        task_result = None
        if len(next_service_keys) == 0 and len(next_messages) == 0:
            # no service component found, return the final result
            last_modules_run = state.get(LAST_MODULES_RUN, [])

            result_dict = run_state.result_outputs[module_key or last_modules_run[-1]]
            if len(result_dict) == 1:
                result = str(next(iter(result_dict.values())))
            else:
                result = str(result_dict)

            task_result = TaskResult(
                task_id=task_def.task_id,
                result=result,
                history=[],
            )

            next_messages.append(
                QueueMessage(
                    type="human",
                    action=ActionTypes.COMPLETED_TASK,
                    data=task_result.model_dump(),
                )
            )

        state[RUN_STATE_KEY] = pickle.dumps(run_state)
        state[NEXT_SERVICE_KEYS] = next_service_keys
        state[RESULT_KEY] = (
            task_result.model_dump() if task_result is not None else None
        )

        return next_messages, state

    async def add_result_to_state(
        self,
        result: TaskResult,
        state: Dict[str, Any],
    ) -> Dict[str, Any]:
        """Add the result of processing a message to the state. Returns the new state."""

        run_state = pickle.loads(state[RUN_STATE_KEY])
        next_service_keys = state[NEXT_SERVICE_KEYS]

        # process the output of the service component(s)
        for module_key in next_service_keys:
            process_component_output(
                self.pipeline,
                run_state,
                module_key,
                result,
            )

        state[RUN_STATE_KEY] = pickle.dumps(run_state)
        state[LAST_MODULES_RUN] = next_service_keys
        return state
